# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "symmray"
copyright = "2022, Johnnie Gray"
author = "Johnnie Gray"

# The full version, including alpha/beta/rc tags
release = "0.0.1"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_nb",
    "sphinx.ext.napoleon",
]

nb_execution_mode = "off"
myst_heading_anchors = 4

# sphinx-autoapi
extensions.append("autoapi.extension")
autoapi_dirs = ["../symmray"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"
# html_theme = 'sphinx_book_theme'

# TODO:
html_theme_options = {
    # "sidebar_hide_name": True,
    # "light_css_variables": {
    #     "color-brand-primary": "hsl(72, 75%, 40%)",
    #     "color-brand-content": "hsl(238, 50%, 60%)",
    # },
    # "dark_css_variables": {
    #     "color-brand-primary": "hsl(72, 75%, 60%)",
    #     "color-brand-content": "hsl(238, 75%, 70%)",
    # },
    # "light_logo": "logo-full.png",
    # "dark_logo": "logo-full.png",
}

# TODO: pygments_style = "lovelace"
# TODO: pygments_dark_style = "stata-dark"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# TODO:
# html_favicon = "_static/logo-favicon.ico"


def setup(app):
    app.add_css_file("my-styles.css")
