@import url('https://fonts.googleapis.com/css2?family=Atkinson+Hyperlegible:ital,wght@0,400;0,700;1,400;1,700&family=IBM+Plex+Mono:ital,wght@0,400;0,600;1,400;1,600&display=swap');

h1, h2, h3 {
    font-family: 'IBM Plex Mono', monospace;
    font-weight: bold;
}

body {
    font-family: 'Atkinson Hyperlegible', sans-serif;
}

code, kbd, pre, samp {
    font-family: 'IBM Plex Mono', monospace;
}

div.cell div.cell_input {
	padding-left: 0em;
	padding-right: 0em;
	border: 1px rgba(127, 127, 127, 0.1) solid;
	background-color: rgba(127, 127, 127, 0.1);
	border-left-color: green;
	border-left-width: medium;
}

.cell_output .output.text_plain, .cell_output .output.traceback, .cell_output .output.stream, .cell_output .output.stderr {
	border: 1px solid rgba(127, 127, 127, 0.0);
    background: rgba(127, 127, 127, 0.0);
	margin-top: 0em;
	margin-bottom: 0em;
	margin-left: 0.5em;
	box-shadow: none;
}

.cell_output {
	padding-left: 1em;
	padding-right: 0em;
	margin-top: 0em;
}
